{  Hi-color-raster-variable-resolution_TV.spin

   Doug Dingus  07/07   Rev 1.0   For HYDRA

  This is a modification to the Easy NTSC driver, first posted by CardboardGuru.  I've
  hardcoded some timing values in order to make selecting resolutions easier.  This
  version of the driver allows variable vertical pixel heights.                    

  The only passed parameter is the beginning of the display bitmap.  Set the pixel
  parameters here, before starting the driver from your spin program.

  Valid horizontal pixel values are:  20, 40, 80, 160, 256, 320

  Valid vertical pixel heights are:  1, 2, 3, 4, etc...

  The last pixel row will be clipped to 192 display lines, therefore, good vertical
  height values are even divisors of 192.

  All display modes consume 1 byte per pixel, resulting in a 125 color display.  This is
  every standard color the Propeller can produce with the reference video circuit,
  without artifacting.  

  At low resolutions, this driver is actually useful.  160x96 only consumes 15K of RAM
  and will produce a useful multi-color display.  Higher resolutions consume most of the
  propeller RAM (30K), with the upper limit (320 x 192) exceeding available RAM space.

  The next version will feature tiles, to take better advantage of the RAM.

  I have the really low resolutions in here for testing overlay sprite type graphics, to
  be running in another COG.

  I've omitted some of CardboardGurus comments for clarity.  See the easy NTSC demo code
  for those areas of this driver init that might not make sense.

  Written for HYDRA -- Doug Dingus   /   CardboardGuru.


}
CON
  ' Set up the processor clock in the standard way for 80MHz
  _CLKMODE = xtal1 + pll16x
  _XINFREQ = 5_000_000

   
PUB start(tvpointer) | i, j, k

 cognew(@entry,tvpointer)
  
DAT                     org
entry                   jmp     #initialization         'Jump past the constants



CON  NTSC_color_frequency       =     3_579_545
DAT  NTSC_color_freq            long  NTSC_color_frequency


CON  NTSC_hsync_clocks          =               624
DAT  NTSC_hsync_VSCL              long  160368
DAT  NTSC_control_signal_palette long  $00_00_02_8a
DAT  NTSC_hsync_pixels          long  %%11_0000_1_2222222_11


'***************************************************
'* Blank lines                                     *
'***************************************************
CON  NTSC_active_video_clocks   =     3008
DAT  NTSC_active_video_VSCL     long  NTSC_active_video_clocks



'***************************************************
'* User graphics lines                             *
'***************************************************
' The important lines at last.  You can edit CHOOSE_horizontal_pixels (40, 80, 160, 256, 320)
' You can also edit CHOOSE_vertical_pixel_height.  (0 - 1)  
CON CHOOSE_horizontal_pixels  = 160
CON CHOOSE_vertical_pixel_height = 2   '1++ for variable vertical pixel heights
DAT vpixel                      long    CHOOSE_vertical_pixel_height
CON CHOOSE_clocks_per_gfx_pixel = 2560 / CHOOSE_horizontal_pixels
CON CALC_bytes_per_line = CHOOSE_horizontal_pixels
CON CALC_waitvids = CHOOSE_horizontal_pixels / 4

'4 pixels per waitvid.  We send their colors each time, for one byte per color.

CON CALC_clocks_per_gfx_frame   =  CHOOSE_clocks_per_gfx_pixel*4



DAT CALC_user_data_VSCL         long  CHOOSE_clocks_per_gfx_pixel << 12 + CALC_clocks_per_gfx_frame


CON CALC_frames_per_gfx_line    = CHOOSE_horizontal_pixels/4

CON CALC_overscan = 448 

CON CHOOSE_horizontal_offset    = 00

CON CALC_backporch = 208 + CHOOSE_horizontal_offset    'this must be a multiple of the total
                                                       'pixel clock 16 clocks in this case.
                                                       'only important if one is artifacting
                                                       'for high-color.  Do what you want
                                                       'here that makes sense otherwise.
CON CALC_frontporch = (CALC_overscan - CALC_backporch)

DAT

'***************************************************
'* The code                                        *
'***************************************************
' Start of real code

'all comments in here still valid, so I left them.

initialization          'set up VCFG

                        ' VCFG: setup Video Configuration register and 3-bit tv DAC pins to output
                        movs    VCFG, #%0111_0000       ' VCFG'S = pinmask (pin31: 0000_0111 : pin24)
                        movd    VCFG, #1                ' VCFG'D = pingroup (grp. 3 i.e. pins 24-31)

                        movi    VCFG, #%0_11_111_000    ' baseband video on bottom nibble, 2-bit color, enable chroma on broadcast & baseband
                                                        ' %0_xx_x_x_x_xxx : Not used
                                                        ' %x_10_x_x_x_xxx : Composite video to top nibble, broadcast to bottom nibble
                                                        ' %x_xx_1_x_x_xxx : 4 color mode
                                                        ' %x_xx_x_1_x_xxx : Enable chroma on broadcast
                                                        ' %x_xx_x_x_1_xxx : Enable chroma on baseband
                                                        ' %x_xx_x_x_x_000 : Broadcast Aural FM bits (don't care)

                        or      DIRA, tvport_mask       ' set DAC pins to output

                        ' 
                        or      DIRA, #1                ' enable debug LED
                        mov     OUTA, #1                ' light up debug LED

                        ' CTRA: setup Frequency to Drive Video
                        movi    CTRA,#%00001_111        ' pll internal routed to Video, PHSx+=FRQx (mode 1) + pll(16x)
                        mov     r1, NTSC_color_freq     ' r1: Color frequency in Hz (3.579_545MHz)
                        rdlong  v_clkfreq, #0           ' copy clk frequency. (80Mhz)
                        mov     r2, v_clkfreq           ' r2: CLKFREQ (80MHz)
                        call    #dividefract            ' perform r3 = 2^32 * r1 / r2
                        mov     v_freq, r3              ' v_freq now contains frqa.       (191)
                        mov     FRQA, r3                ' set frequency for counter

                                                
                        mov     A, PAR                  'get bitmap address from PAR register
                        mov     bmp, PAR                'store it elsewhere as A will be written to

                        

'-----------------------------------------------------------------------------
                        'NTSC has 244 "visible" lines, but some of that is overscan etc.
                        '  so pick a number of lines for user graphics e.g. 192
                        '  then display (244-192)/2 = 26 lines of overscan before
                        '  and after the user display.
frame_loop
                        mov     line_loop, #26          '(26 so far)
                        'There's always a horizontal sync at the start of each line.  The clever stuff is
                        '  in the constants described earlier.
:vert_back_porch        mov     VSCL, NTSC_hsync_VSCL
                        waitvid NTSC_control_signal_palette, hsync
                        
                        'This is just a blank line.  Notice how the whole line, apart from the
                        '  hsyc is output with one WAITVID.
                        
                        mov     VSCL, NTSC_active_video_VSCL
                        waitvid brown_border_in_color0, #0
                        djnz    line_loop, #:vert_back_porch
'-----------------------------------------------------------------------------
                        'Time to do the user graphics lines.  We're having 192 of them.
                        mov     line_loop, #192         '(218 so far)
                        mov     vpixel_cnt, vpixel
                        'hsync                        
user_graphics_lines     mov     VSCL, NTSC_hsync_VSCL
                        waitvid NTSC_control_signal_palette, hsync
                        'The overscan before the user graphics.   
                        mov     VSCL,#CALC_backporch
                        waitvid green_border_in_color0, #0 
                        
                        'And now at long last, the user graphics.  Program the VSCL for
                        '  4 pixels per frame.  
                        mov     VSCL, CALC_user_data_VSCL
                        

                        'this is where the bitmap gets drawn.  We load a counter with the
                        'number of waitvids / line, and loop until they are done.

                        mov     r1, #CALC_waitvids  'number of horiz pixels / 4
:draw_pixels            rdlong  B, A    'get four pixels from hub

                        {

                        not sure how to do CLUT translation here yet.  I do know
                        at 320 pixels, there is only room for maybe two opcodes
                        lower, say 160 and below, there is room for several

                        }
                        
                        waitvid B, #%%3210  'draw them to screen
                        add     A, #4    'point to next pixel group
                        djnz    r1, #:draw_pixels  'line done?


                        'We've finished drawing a user graphics line, whether it was color
                        '  bar or flag.  Now do the overscan bit.  
end_of_flag_line        mov    VSCL,#CALC_frontporch
                        waitvid blue_border_in_color0 , border

                        'here is where we deal with vertical height.  Things are either
                        'one scanline high, or two.  So and the line_loop counter with
                        'a binary 1 to sort out even from odd lines and subtract
                        'from the bitmap memory pointer depending.

                        sub     vpixel_cnt, #1  wz
               if_z     mov     vpixel_cnt, vpixel
               if_nz    sub     A, #CALC_bytes_per_line         

                        
                        'keep doing the 192 active display area
                        djnz    line_loop, #user_graphics_lines


                        'Reset raster display pointer for next frame
                        mov     A, bmp
'-----------------------------------------------------------------------------
                        'Overscan at the bottom of screen.  
                        mov     line_loop, #26          '(244)
                        'hsync
vert_front_porch        mov     VSCL, NTSC_hsync_VSCL
                        waitvid NTSC_control_signal_palette, hsync
                        mov     VSCL, NTSC_active_video_VSCL
                        waitvid magenta_border_in_color0, #0
                        djnz    line_loop, #vert_front_porch
'-----------------------------------------------------------------------------
                        'This is the vertical sync.  It consists of 3 sections of 6 lines each.
                        mov     line_loop, #6           '(250)
:vsync_higha            mov     VSCL, NTSC_hsync_VSCL
                        waitvid NTSC_control_signal_palette, vsync_high_1
                        mov     VSCL, NTSC_active_video_VSCL
                        waitvid NTSC_control_signal_palette, vsync_high_2
                        djnz    line_loop, #:vsync_higha
'-----------------------------------------------------------------------------
                        mov     line_loop, #6           '(256)
:vsync_low              mov     VSCL, NTSC_hsync_VSCL
                        waitvid NTSC_control_signal_palette, vsync_low_1
                        mov     VSCL, NTSC_active_video_VSCL
                        waitvid NTSC_control_signal_palette, vsync_low_2
                        djnz    line_loop, #:vsync_low
'-----------------------------------------------------------------------------
                        mov     line_loop, #6           '(250)
:vsync_highb            mov     VSCL, NTSC_hsync_VSCL
                        waitvid NTSC_control_signal_palette, vsync_high_1
                        mov     VSCL, NTSC_active_video_VSCL
                        waitvid NTSC_control_signal_palette, vsync_high_2
                        djnz    line_loop, #:vsync_highb
'-----------------------------------------------------------------------------
                        mov     OUTA, #0

                        jmp     #frame_loop






' /////////////////////////////////////////////////////////////////////////////
' dividefract:
' Perform 2^32 * r1/r2, result stored in r3 (useful for TV calc)
' This is taken from the tv driver.
' NOTE: It divides a bottom heavy fraction e.g. 1/2 and gives the result as a 32-bit fraction.
' /////////////////////////////////////////////////////////////////////////////
dividefract                                     
                        mov     r0,#32+1
:loop                   cmpsub  r1,r2           wc
                        rcl     r3,#1
                        shl     r1,#1
                        djnz    r0,#:loop

dividefract_ret         ret                             '+140


' General Purpose Registers (un initialized)
r0                      long                    $0    ' should typically equal 0
r1                      long                    $0
r2                      long                    $0
r3                      long                    $0

A                       long                    $0  'coupla more general purpose registers
B                       long                    $0
bmp                     long                    $0  'tvpointer ends up here
C                       long                    $0
vpixel_cnt              long                    $0  'counter for vertical pixel height

'This is the scan line counter
line_loop               long    0




' Video (TV) Registers
tvport_mask             long                    %0000_0111<<12

v_freq                  long                    0

' Graphics related vars.
v_coffset               long                    $02020202  ' color offset (every color is added by $02)
v_clkfreq               long                    $0




'Initialized data below here


'Pixel streams
'  These are shifted out of the VSU to the right, so lowest bits are actioned first.
'
hsync                   long    %%11_0000_1_2222222_11  ' Used with NTSC_control_signal_palette so:
                                                        '      0 = blanking level
                                                        '      1 = Black
                                                        '      2 = Color NTSC_control_signal_palette (yellow at zero value)
vsync_high_1            long    %%11111111111_222_11
vsync_high_2            long    %%1111111111111111
vsync_low_1             long    %%22222222222222_11
vsync_low_2             long    %%1_222222222222222
all_black               long    %%1111111111111111
border                  long    %%0000000000000000


'Palettes
'  These are always 4 colors (or blanking level) stored in reverse order:
'                               Color3_Color2_Color1_Color0

blue_border_in_color0   long    $03030303     'All set to grey
green_border_in_color0  long    $03030303     'Feel free to change these for interesting 
magenta_border_in_color0 long   $03030303     'border colors
brown_border_in_color0  long    $03030303



'Big Lookup table because I've got room right now for the CLUT.  Not implemented right now.
'CLUT being done in higher level program.  

clut   long $02020202    'Black = 00
       long $03030303
       long $04040404
       long $05050505
       long $06060606    'White = 100% IRE
       long $07070707
       long $07070707
       long $07070707    'Duped for lack of precision in DAC

       
       long $19191919    'Each of the 16 hues possible follows
       long $1a1a1a1a
       long $1b1b1b1b    'With the high-saturation colors mapped
       long $1c1c1c1c    'where they make sense
       long $1d1d1d1d
       long $1e1e1e1e
       long $98989898
       long $afafafaf

       
       long $29292929
       long $2a2a2a2a
       long $2b2b2b2b
       long $2c2c2c2c
       long $2d2d2d2d
       long $2e2e2e2e
       long $a8a8a8a8
       long $bfbfbfbf

       
       long $39393939
       long $3a3a3a3a
       long $3b3b3b3b
       long $3c3c3c3c
       long $3d3d3d3d
       long $3e3e3e3e
       long $b8b8b8b8
       long $cfcfcfcf

       
       long $49494949
       long $4a4a4a4a
       long $4b4b4b4b
       long $4c4c4c4c
       long $4d4d4d4d
       long $4e4e4e4e
       long $c8c8c8c8
       long $dfdfdfdf
       
       long $59595959
       long $5a5a5a5a
       long $5b5b5b5b
       long $5c5c5c5c
       long $5d5d5d5d
       long $5e5e5e5e
       long $d8d8d8d8
       long $efefefef
       
       long $69696969
       long $6a6a6a6a
       long $6b6b6b6b
       long $6c6c6c6c
       long $6d6d6d6d
       long $6e6e6e6e
       long $e8e8e8e8
       long $ffffffff

       long $79797979
       long $7a7a7a7a
       long $7b7b7b7b
       long $7c7c7c7c
       long $7d7d7d7d
       long $7e7e7e7e
       long $f8f8f8f8
       long $0f0f0f0f
       
       long $89898989
       long $8a8a8a8a
       long $8b8b8b8b
       long $8c8c8c8c
       long $8d8d8d8d
       long $8e8e8e8e
       long $08080808
       long $1f1f1f1f
       
       long $99999999
       long $9a9a9a9a
       long $9b9b9b9b
       long $9c9c9c9c
       long $9d9d9d9d
       long $9e9e9e9e
       long $18181818
       long $2f2f2f2f
       
       long $a9a9a9a9
       long $aaaaaaaa
       long $abababab
       long $acacacac
       long $adadadad
       long $aeaeaeae
       long $28282828
       long $3f3f3f3f

       long $b9b9b9b9
       long $babababa
       long $bbbbbbbb
       long $bcbcbcbc
       long $bdbdbdbd
       long $bebebebe
       long $38383838
       long $4f4f4f4f
       
       long $c9c9c9c9
       long $cacacaca
       long $cbcbcbcb
       long $cccccccc
       long $cdcdcdcd
       long $cececece
       long $48484848
       long $5f5f5f5f
       
       long $d9d9d9d9
       long $dadadada
       long $dbdbdbdb
       long $dcdcdcdc
       long $dddddddd
       long $dededede
       long $58585858
       long $6f6f6f6f
           
       long $e9e9e9e9
       long $eaeaeaea
       long $ebebebeb
       long $ecececec
       long $edededed
       long $eeeeeeee
       long $68686868
       long $7f7f7f7f
       
       long $f9f9f9f9
       long $fafafafa
       long $fbfbfbfb
       long $fcfcfcfc
       long $fdfdfdfd
       long $fefefefe
       long $78787878
       long $8f8f8f8f

                                   